/* -*- mode: c++ -*- */
/**
 * @file   Shape.h
 * @author Sebastien Fourey <http://www.greyc.ensicaen.fr/~seb>
 * @date   Sat Aug 18 2007
 * 
 * @brief  
 * @copyright
 * Copyright or  or Copr. Sbastien Fourey
 * 
 * Sbastien.Fourey - http://www.greyc.ensicaen.fr/~seb
 * 
 * 
 * This source code is part of the Board project, a C++ library whose
 * purpose is to allow simple drawings in EPS, FIG or SVG files.
 * 
 * This software is governed by the CeCILL-C license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-C
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms
 * 
 */
#ifndef _BOARD_SHAPES_H_
#define _BOARD_SHAPES_H_

#include "board/Point.h"
#include "board/Rect.h"
#include "board/Color.h"
#include "board/Transforms.h"
#include <string>
#include <vector>
#include <iostream>
#include <map>

namespace BoardLib {

/**
 * Shape structure.
 * @brief Abstract structure for a 2D shape.
 */
struct Shape {

  enum LineCap { ButtCap = 0, RoundCap, SquareCap };
  enum LineJoin { MiterJoin = 0, RoundJoin, BevelJoin };

  unsigned int depth;		/**< The depth of the shape. */
  Color penColor;		/**< The color of the shape. */
  Color fillColor;		/**< The color of the shape. */
  float lineWidth;		/**< The line thickness. */
  LineCap lineCap;		/**< The linecap attribute. (The way line terminates.) */
  LineJoin lineJoin;		/**< The linejoin attribute. (The shape of line junctions.) */

  /** 
   * Shape constructor.
   * 
   * @param penColor The pen color of the shape.
   * @param fillColor The fill color of the shape.
   * @param lineWidth The line thickness.
   * @param depth The depth of the shape.
   */
  Shape( Color penColor, Color fillColor,
	 float lineWidth, const LineCap cap, const LineJoin join,
	 unsigned int depth )
    : depth( depth ), penColor( penColor ), fillColor( fillColor ), 
      lineWidth( lineWidth ), lineCap( cap ), lineJoin( join ) { }

  /** 
   * Shape destructor.
   */
  virtual ~Shape() { }
  
  /** 
   * Checks whether a shape is filled with a color or not.
   * 
   * @return true if the shape is filled.
   */
  inline bool filled() const { return fillColor != Color::none; }

  /** 
   * Writes the EPS code of the shape in a stream according
   * to a transform.
   * 
   * @param stream The output stream.
   * @param transform A 2D transform to be applied.
   */
  virtual void flushPostscript( std::ostream & stream,
				const TransformEPS & transform ) const = 0;

  /** 
   * Writes the FIG code of the shape in a stream according
   * to a transform.
   * 
   * @param stream The output stream.
   * @param transform A 2D transform to be applied.
   */
  virtual void flushFIG( std::ostream & stream,
			 const TransformFIG & transform,
			 std::map<Color,int> & colormap ) const = 0;

  /** 
   * Writes the SVG code of the shape in a stream according
   * to a transform.
   * 
   * @param stream The output stream.
   * @param transform A 2D transform to be applied.
   */
  virtual void flushSVG( std::ostream & stream,
			 const TransformSVG & transform ) const = 0;
  
  /** 
   * Returns the bounding box of the figure.
   *
   * @return The rectangle of the bounding box.
   */
  virtual Rect boundingBox() const = 0;

protected:

  /** 
   * Return a string of the svg properties lineWidth, opacity, penColor, fillColor,
   * lineCap, and lineJoin.
   * 
   * @return A string of the properties suitable for inclusion in an svg tag.
   */
  std::string svgProperties( const TransformSVG & transform ) const;
};

/**
 * The line structure.
 * @brief A line between two points.
 */
struct Line : public Shape { 
  double x1;			/**< First coordinate of the start point. */
  double y1;			/**< Second coordinate of the start point. */
  double x2; 			/**< First coordinate of the end point. */
  double y2;			/**< Second coordinate of the end point. */
  
  /** 
   * Constructs a line.
   * 
   * @param x1 First coordinate of the start point.
   * @param y1 Second coordinate of the start point.
   * @param x2 First coordinate of the end point.
   * @param y2 Second coordinate of the end point.
   * @param color The color of the line.
   * @param lineWidth The line thickness.
   * @param depth The depth of the line.
   */
  Line( double x1, double y1, double x2, double y2, 
	Color color, 
	float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	unsigned int depth = 0 )
    : Shape( color, Color::none, lineWidth, cap, join, depth ),
      x1( x1 ), y1( y1 ), x2( x2 ), y2( y2 ) { }

  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;
  
  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};

/**
 * The arrow structure.
 * @brief A line between two points with an arrow at one extremity.
 */
struct Arrow : public Line { 

  /** 
   * Constructs an arrow.
   * 
   * @param x1 First coordinate of the start point.
   * @param y1 Second coordinate of the start point.
   * @param x2 First coordinate of the end point.
   * @param y2 Second coordinate of the end point.
   * @param penColor The color of the line.
   * @param fillColor The fill color of the sharp end.
   * @param lineWidth The line thickness.
   * @param depth The depth of the line.
   */
  Arrow( double x1, double y1, double x2, double y2,
	 Color penColor, Color fillColor,
	 float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	 unsigned int depth = 0 )
    : Line( x1, y1, x2, y2, penColor, lineWidth, cap, join, depth ) {
    Shape::fillColor = fillColor;
  }
  
  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;
  
  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;
  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

};

/**
 * The polyline structure.
 * @brief A polygonal line described by a series of 2D points.
 */
struct Polyline : public Shape { 
  std::vector<Point> points;
  bool closed;

  Polyline( const std::vector<Point> & points, 
	    bool closed,
	    Color penColor, Color fillColor,
	    float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	    unsigned int depth = 0 )
    : Shape( penColor, fillColor, lineWidth, cap, join, depth ),
      points( points ), closed( closed ) { }
	    
  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};


/**
 * The GouraudTriangle structure.
 * @brief A triangle with shaded filling according to colors given for each vertex. 
 */
struct GouraudTriangle : public Polyline {
  Color color0;
  Color color1;
  Color color2;
  int subdivisions;

  GouraudTriangle( const Point & p0, const Color & color0,
		   const Point & p1, const Color & color1,
		   const Point & p2, const Color & color2,
		   int subdivisions,
		   unsigned int depth = 0 );

  GouraudTriangle( const Point & p0, float brightness0,
		   const Point & p1, float brightness1,
		   const Point & p2, float brightness2,
		   const Color & fillColor,
		   int subdivisions,
		   unsigned int depth = 0 );

  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  /** 
   * Sends the Triangle to a FIG file format stream.
   * <p><b>Warning!</b> Because shading would generally require
   * more colors in the colormap than allowed by the FIG file format, 
   * rendering a Gouraud triangle in an XFig file is the same as rendering
   * a simple triangle whose filling color is the average of the vertex colors.
   * 
   * @param stream 
   * @param transform 
   * @param Color 
   * @param colormap 
   */
  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

};

/**
 * The rectangle structure.
 * @brief A rectangle.
 */
struct Rectangle : public Shape {
  double x;
  double y;
  double width;
  double height;

  Rectangle( double x, double y, double width, double height,
	     Color penColor, Color fillColor,
	     float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	     unsigned int depth = 0 )
    : Shape( penColor, fillColor, lineWidth, cap, join, depth ),
      x( x ), y( y ), width( width ), height( height ) { }

  Rectangle( const Rect & rect,
	     Color penColor, Color fillColor,
	     float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	     unsigned int depth = 0 )
    : Shape( penColor, fillColor, lineWidth, cap, join, depth ),
      x( rect.left ), y( rect.top ), width( rect.width ), height( rect.height ) { }
  
  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};

/**
 * The circle structure.
 * @brief A circle.
 */
struct Circle : public Shape {
  double x;
  double y;
  double radius;
  
  Circle( double x, double y, double radius, 
	  Color penColor, Color fillColor,
	  float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	  unsigned int depth = 0 )
    : Shape( penColor, fillColor, lineWidth, cap, join, depth ),
      x( x ), y( y ), radius( radius ) { }

  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};

/**
 * The ellipse structure.
 * @brief An ellipse.
 */
struct Ellipse : public Shape {
  double x;
  double y;
  double xRadius; 
  double yRadius;
  
  Ellipse( double x, double y, 
	   double xRadius, double yRadius, 
	   Color penColor, Color fillColor,
	   float lineWidth, const LineCap cap = ButtCap, const LineJoin join = MiterJoin,
	   unsigned int depth = 0 )
    : Shape( penColor, fillColor, lineWidth, cap, join, depth ),
      x( x ), y( y ), xRadius( xRadius ), yRadius( yRadius ) { }

  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};

/**
 * The text structure.
 * @brief A piece of text.
 */
struct Text : public Shape {
  float x;
  float y;
  std::string text;
  std::string font;
  float size;
  
  Text( float x, float y,
	const std::string & text, const std::string & font, float size,
	Color color, unsigned int depth )
    : Shape( color, Color::none, 1.0, ButtCap, MiterJoin, depth ),
      x( x ), y( y ), text( text ), font( font ), size( size ) { }
  
  void flushPostscript( std::ostream & stream,
			const TransformEPS & transform ) const;

  void flushFIG( std::ostream & stream,
		 const TransformFIG & transform,
		 std::map<Color,int> & colormap ) const;

  void flushSVG( std::ostream & stream,
		 const TransformSVG & transform ) const;

  Rect boundingBox() const;
};

/** 
 * Compares two shapes according to their depths.
 * 
 * @param s1 A pointer to a first shape.
 * @param s2 A pointer to a second shape.
 * 
 * @return 
 */
bool shapeGreaterDepth( const Shape *s1, const Shape *s2 );

} // namespace BoardLib 

#endif /* _SHAPE_H_ */

